from google.protobuf import any_pb2 as _any_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from qm.grpc.qm.grpc.v2 import common_types_pb2 as _common_types_pb2
from qm.grpc.qm.pb import compiler_pb2 as _compiler_pb2
from qm.grpc.qm.pb import frontend_pb2 as _frontend_pb2
from qm.grpc.qm.pb import inc_qua_pb2 as _inc_qua_pb2
from qm.grpc.qm.pb import inc_qua_config_pb2 as _inc_qua_config_pb2
from qm.grpc.qm.pb import qm_manager_pb2 as _qm_manager_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetVersionRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class QmmServiceResetDataProcessingRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class QmmServiceGetQuantumMachineRequest(_message.Message):
    __slots__ = ("machine_id",)
    MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    machine_id: str
    def __init__(self, machine_id: _Optional[str] = ...) -> None: ...

class QmmServiceCloseQuantumMachineRequest(_message.Message):
    __slots__ = ("machine_id",)
    MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
    machine_id: str
    def __init__(self, machine_id: _Optional[str] = ...) -> None: ...

class HealthCheckRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class HealthCheckResponse(_message.Message):
    __slots__ = ("success", "error")
    class HealthCheckResponseSuccess(_message.Message):
        __slots__ = ("ok", "components", "details")
        class ComponentsEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: HealthCheckResponse.HealthCheckResponseSuccess
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[HealthCheckResponse.HealthCheckResponseSuccess, _Mapping]] = ...) -> None: ...
        class DetailsEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: _any_pb2.Any
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_any_pb2.Any, _Mapping]] = ...) -> None: ...
        OK_FIELD_NUMBER: _ClassVar[int]
        COMPONENTS_FIELD_NUMBER: _ClassVar[int]
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        ok: bool
        components: _containers.MessageMap[str, HealthCheckResponse.HealthCheckResponseSuccess]
        details: _containers.MessageMap[str, _any_pb2.Any]
        def __init__(self, ok: bool = ..., components: _Optional[_Mapping[str, HealthCheckResponse.HealthCheckResponseSuccess]] = ..., details: _Optional[_Mapping[str, _any_pb2.Any]] = ...) -> None: ...
    class HealthCheckResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: HealthCheckResponse.HealthCheckResponseSuccess
    error: HealthCheckResponse.HealthCheckResponseError
    def __init__(self, success: _Optional[_Union[HealthCheckResponse.HealthCheckResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[HealthCheckResponse.HealthCheckResponseError, _Mapping]] = ...) -> None: ...

class OpenQuantumMachineRequest(_message.Message):
    __slots__ = ("config", "close_mode")
    class CloseMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        CLOSE_MODE_UNSPECIFIED: _ClassVar[OpenQuantumMachineRequest.CloseMode]
        CLOSE_MODE_IF_NEEDED: _ClassVar[OpenQuantumMachineRequest.CloseMode]
        CLOSE_MODE_ALL: _ClassVar[OpenQuantumMachineRequest.CloseMode]
    CLOSE_MODE_UNSPECIFIED: OpenQuantumMachineRequest.CloseMode
    CLOSE_MODE_IF_NEEDED: OpenQuantumMachineRequest.CloseMode
    CLOSE_MODE_ALL: OpenQuantumMachineRequest.CloseMode
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    CLOSE_MODE_FIELD_NUMBER: _ClassVar[int]
    config: _inc_qua_config_pb2.QuaConfig
    close_mode: OpenQuantumMachineRequest.CloseMode
    def __init__(self, config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ..., close_mode: _Optional[_Union[OpenQuantumMachineRequest.CloseMode, str]] = ...) -> None: ...

class OpenQuantumMachineResponse(_message.Message):
    __slots__ = ("success", "error")
    class OpenQuantumMachineResponseSuccess(_message.Message):
        __slots__ = ("quantum_machine_id", "open_qm_warnings")
        QUANTUM_MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
        OPEN_QM_WARNINGS_FIELD_NUMBER: _ClassVar[int]
        quantum_machine_id: str
        open_qm_warnings: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.OpenQmWarning]
        def __init__(self, quantum_machine_id: _Optional[str] = ..., open_qm_warnings: _Optional[_Iterable[_Union[_qm_manager_pb2.OpenQmWarning, _Mapping]]] = ...) -> None: ...
    class OpenQuantumMachineResponseError(_message.Message):
        __slots__ = ("config_validation_errors", "physical_validation_errors")
        CONFIG_VALIDATION_ERRORS_FIELD_NUMBER: _ClassVar[int]
        PHYSICAL_VALIDATION_ERRORS_FIELD_NUMBER: _ClassVar[int]
        config_validation_errors: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.ConfigValidationMessage]
        physical_validation_errors: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.PhysicalValidationMessage]
        def __init__(self, config_validation_errors: _Optional[_Iterable[_Union[_qm_manager_pb2.ConfigValidationMessage, _Mapping]]] = ..., physical_validation_errors: _Optional[_Iterable[_Union[_qm_manager_pb2.PhysicalValidationMessage, _Mapping]]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: OpenQuantumMachineResponse.OpenQuantumMachineResponseSuccess
    error: OpenQuantumMachineResponse.OpenQuantumMachineResponseError
    def __init__(self, success: _Optional[_Union[OpenQuantumMachineResponse.OpenQuantumMachineResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[OpenQuantumMachineResponse.OpenQuantumMachineResponseError, _Mapping]] = ...) -> None: ...

class GetQuantumMachineResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetQuantumMachineResponseSuccess(_message.Message):
        __slots__ = ("machine_id", "config")
        MACHINE_ID_FIELD_NUMBER: _ClassVar[int]
        CONFIG_FIELD_NUMBER: _ClassVar[int]
        machine_id: str
        config: _inc_qua_config_pb2.QuaConfig
        def __init__(self, machine_id: _Optional[str] = ..., config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ...) -> None: ...
    class GetQuantumMachineResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetQuantumMachineResponse.GetQuantumMachineResponseSuccess
    error: GetQuantumMachineResponse.GetQuantumMachineResponseError
    def __init__(self, success: _Optional[_Union[GetQuantumMachineResponse.GetQuantumMachineResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetQuantumMachineResponse.GetQuantumMachineResponseError, _Mapping]] = ...) -> None: ...

class GetControllersRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListOpenQuantumMachinesRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListOpenQuantumMachinesResponse(_message.Message):
    __slots__ = ("success", "error")
    class ListOpenQuantumMachinesResponseSuccess(_message.Message):
        __slots__ = ("machine_ids",)
        MACHINE_IDS_FIELD_NUMBER: _ClassVar[int]
        machine_ids: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, machine_ids: _Optional[_Iterable[str]] = ...) -> None: ...
    class ListOpenQuantumMachinesResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: ListOpenQuantumMachinesResponse.ListOpenQuantumMachinesResponseSuccess
    error: ListOpenQuantumMachinesResponse.ListOpenQuantumMachinesResponseError
    def __init__(self, success: _Optional[_Union[ListOpenQuantumMachinesResponse.ListOpenQuantumMachinesResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[ListOpenQuantumMachinesResponse.ListOpenQuantumMachinesResponseError, _Mapping]] = ...) -> None: ...

class CloseAllQuantumMachinesRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class JobsQueryParams(_message.Message):
    __slots__ = ("quantum_machine_ids", "job_ids", "user_ids", "description", "status")
    QUANTUM_MACHINE_IDS_FIELD_NUMBER: _ClassVar[int]
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    USER_IDS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    quantum_machine_ids: _containers.RepeatedScalarFieldContainer[str]
    job_ids: _containers.RepeatedScalarFieldContainer[str]
    user_ids: _containers.RepeatedScalarFieldContainer[str]
    description: str
    status: _containers.RepeatedScalarFieldContainer[_common_types_pb2.JobExecutionStatus]
    def __init__(self, quantum_machine_ids: _Optional[_Iterable[str]] = ..., job_ids: _Optional[_Iterable[str]] = ..., user_ids: _Optional[_Iterable[str]] = ..., description: _Optional[str] = ..., status: _Optional[_Iterable[_Union[_common_types_pb2.JobExecutionStatus, str]]] = ...) -> None: ...

class JobResponseData(_message.Message):
    __slots__ = ("job_id", "description", "status", "metadata", "is_simulation")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    IS_SIMULATION_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    description: str
    status: _common_types_pb2.JobExecutionStatus
    metadata: JobMetadata
    is_simulation: bool
    def __init__(self, job_id: _Optional[str] = ..., description: _Optional[str] = ..., status: _Optional[_Union[_common_types_pb2.JobExecutionStatus, str]] = ..., metadata: _Optional[_Union[JobMetadata, _Mapping]] = ..., is_simulation: bool = ...) -> None: ...

class JobMetadata(_message.Message):
    __slots__ = ("created_at", "started_at", "last_status_updated_at")
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    STARTED_AT_FIELD_NUMBER: _ClassVar[int]
    LAST_STATUS_UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    created_at: _timestamp_pb2.Timestamp
    started_at: _timestamp_pb2.Timestamp
    last_status_updated_at: _timestamp_pb2.Timestamp
    def __init__(self, created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., started_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., last_status_updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class GetJobsRequest(_message.Message):
    __slots__ = ("query",)
    QUERY_FIELD_NUMBER: _ClassVar[int]
    query: JobsQueryParams
    def __init__(self, query: _Optional[_Union[JobsQueryParams, _Mapping]] = ...) -> None: ...

class GetJobsResponse(_message.Message):
    __slots__ = ("success", "error")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetJobsSuccess
    error: GetJobsError
    def __init__(self, success: _Optional[_Union[GetJobsSuccess, _Mapping]] = ..., error: _Optional[_Union[GetJobsError, _Mapping]] = ...) -> None: ...

class GetJobsError(_message.Message):
    __slots__ = ("details",)
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    details: str
    def __init__(self, details: _Optional[str] = ...) -> None: ...

class GetJobsSuccess(_message.Message):
    __slots__ = ("jobs",)
    JOBS_FIELD_NUMBER: _ClassVar[int]
    jobs: _containers.RepeatedCompositeFieldContainer[JobResponseData]
    def __init__(self, jobs: _Optional[_Iterable[_Union[JobResponseData, _Mapping]]] = ...) -> None: ...

class GetControllersResponse(_message.Message):
    __slots__ = ("success", "error")
    class FemType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        FEM_TYPE_UNSPECIFIED: _ClassVar[GetControllersResponse.FemType]
        FEM_TYPE_LF: _ClassVar[GetControllersResponse.FemType]
        FEM_TYPE_MW: _ClassVar[GetControllersResponse.FemType]
    FEM_TYPE_UNSPECIFIED: GetControllersResponse.FemType
    FEM_TYPE_LF: GetControllersResponse.FemType
    FEM_TYPE_MW: GetControllersResponse.FemType
    class ControllerType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        CONTROLLER_TYPE_UNSPECIFIED: _ClassVar[GetControllersResponse.ControllerType]
        CONTROLLER_TYPE_OPX_1000: _ClassVar[GetControllersResponse.ControllerType]
    CONTROLLER_TYPE_UNSPECIFIED: GetControllersResponse.ControllerType
    CONTROLLER_TYPE_OPX_1000: GetControllersResponse.ControllerType
    class GetControllersResponseSuccess(_message.Message):
        __slots__ = ("control_devices",)
        class ControlDevicesEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: GetControllersResponse.Controller
            def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[GetControllersResponse.Controller, _Mapping]] = ...) -> None: ...
        CONTROL_DEVICES_FIELD_NUMBER: _ClassVar[int]
        control_devices: _containers.MessageMap[str, GetControllersResponse.Controller]
        def __init__(self, control_devices: _Optional[_Mapping[str, GetControllersResponse.Controller]] = ...) -> None: ...
    class GetControllersResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    class Controller(_message.Message):
        __slots__ = ("hostname", "controller_type", "fems", "temperatures")
        class FemsEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: int
            value: GetControllersResponse.Fem
            def __init__(self, key: _Optional[int] = ..., value: _Optional[_Union[GetControllersResponse.Fem, _Mapping]] = ...) -> None: ...
        class TemperaturesEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: float
            def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
        HOSTNAME_FIELD_NUMBER: _ClassVar[int]
        CONTROLLER_TYPE_FIELD_NUMBER: _ClassVar[int]
        FEMS_FIELD_NUMBER: _ClassVar[int]
        TEMPERATURES_FIELD_NUMBER: _ClassVar[int]
        hostname: str
        controller_type: GetControllersResponse.ControllerType
        fems: _containers.MessageMap[int, GetControllersResponse.Fem]
        temperatures: _containers.ScalarMap[str, float]
        def __init__(self, hostname: _Optional[str] = ..., controller_type: _Optional[_Union[GetControllersResponse.ControllerType, str]] = ..., fems: _Optional[_Mapping[int, GetControllersResponse.Fem]] = ..., temperatures: _Optional[_Mapping[str, float]] = ...) -> None: ...
    class Fem(_message.Message):
        __slots__ = ("type",)
        TYPE_FIELD_NUMBER: _ClassVar[int]
        type: GetControllersResponse.FemType
        def __init__(self, type: _Optional[_Union[GetControllersResponse.FemType, str]] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetControllersResponse.GetControllersResponseSuccess
    error: GetControllersResponse.GetControllersResponseError
    def __init__(self, success: _Optional[_Union[GetControllersResponse.GetControllersResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetControllersResponse.GetControllersResponseError, _Mapping]] = ...) -> None: ...

class GetVersionResponse(_message.Message):
    __slots__ = ("success", "error")
    class GetVersionResponseSuccess(_message.Message):
        __slots__ = ("gateway", "controllers")
        class ControllersEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        GATEWAY_FIELD_NUMBER: _ClassVar[int]
        CONTROLLERS_FIELD_NUMBER: _ClassVar[int]
        gateway: str
        controllers: _containers.ScalarMap[str, str]
        def __init__(self, gateway: _Optional[str] = ..., controllers: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class GetVersionResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: GetVersionResponse.GetVersionResponseSuccess
    error: GetVersionResponse.GetVersionResponseError
    def __init__(self, success: _Optional[_Union[GetVersionResponse.GetVersionResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[GetVersionResponse.GetVersionResponseError, _Mapping]] = ...) -> None: ...

class QmmServiceSimulateRequest(_message.Message):
    __slots__ = ("config", "high_level_program", "simulate", "controller_connections")
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    HIGH_LEVEL_PROGRAM_FIELD_NUMBER: _ClassVar[int]
    SIMULATE_FIELD_NUMBER: _ClassVar[int]
    CONTROLLER_CONNECTIONS_FIELD_NUMBER: _ClassVar[int]
    config: _inc_qua_config_pb2.QuaConfig
    high_level_program: _inc_qua_pb2.QuaProgram
    simulate: _frontend_pb2.ExecutionRequest.Simulate
    controller_connections: _containers.RepeatedCompositeFieldContainer[_frontend_pb2.InterOpxConnection]
    def __init__(self, config: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ..., high_level_program: _Optional[_Union[_inc_qua_pb2.QuaProgram, _Mapping]] = ..., simulate: _Optional[_Union[_frontend_pb2.ExecutionRequest.Simulate, _Mapping]] = ..., controller_connections: _Optional[_Iterable[_Union[_frontend_pb2.InterOpxConnection, _Mapping]]] = ...) -> None: ...

class QmmServiceSimulateResponse(_message.Message):
    __slots__ = ("success", "error")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: SimulationSuccess
    error: SimulationError
    def __init__(self, success: _Optional[_Union[SimulationSuccess, _Mapping]] = ..., error: _Optional[_Union[SimulationError, _Mapping]] = ...) -> None: ...

class SimulationSuccess(_message.Message):
    __slots__ = ("job_id", "messages", "simulated")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    SIMULATED_FIELD_NUMBER: _ClassVar[int]
    job_id: str
    messages: _containers.RepeatedCompositeFieldContainer[_compiler_pb2.CompilerMessage]
    simulated: _frontend_pb2.SimulatedResponsePart
    def __init__(self, job_id: _Optional[str] = ..., messages: _Optional[_Iterable[_Union[_compiler_pb2.CompilerMessage, _Mapping]]] = ..., simulated: _Optional[_Union[_frontend_pb2.SimulatedResponsePart, _Mapping]] = ...) -> None: ...

class SimulationError(_message.Message):
    __slots__ = ("details", "config_validation_errors", "physical_validation_errors", "messages")
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_VALIDATION_ERRORS_FIELD_NUMBER: _ClassVar[int]
    PHYSICAL_VALIDATION_ERRORS_FIELD_NUMBER: _ClassVar[int]
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    details: str
    config_validation_errors: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.ConfigValidationMessage]
    physical_validation_errors: _containers.RepeatedCompositeFieldContainer[_qm_manager_pb2.PhysicalValidationMessage]
    messages: _containers.RepeatedCompositeFieldContainer[_compiler_pb2.CompilerMessage]
    def __init__(self, details: _Optional[str] = ..., config_validation_errors: _Optional[_Iterable[_Union[_qm_manager_pb2.ConfigValidationMessage, _Mapping]]] = ..., physical_validation_errors: _Optional[_Iterable[_Union[_qm_manager_pb2.PhysicalValidationMessage, _Mapping]]] = ..., messages: _Optional[_Iterable[_Union[_compiler_pb2.CompilerMessage, _Mapping]]] = ...) -> None: ...

class ResetDataProcessingResponse(_message.Message):
    __slots__ = ("success", "error")
    class ResetDataProcessingResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class ResetDataProcessingResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: ResetDataProcessingResponse.ResetDataProcessingResponseSuccess
    error: ResetDataProcessingResponse.ResetDataProcessingResponseError
    def __init__(self, success: _Optional[_Union[ResetDataProcessingResponse.ResetDataProcessingResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[ResetDataProcessingResponse.ResetDataProcessingResponseError, _Mapping]] = ...) -> None: ...

class CloseQuantumMachineResponse(_message.Message):
    __slots__ = ("success", "error")
    class CloseQuantumMachineResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class CloseQuantumMachineResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: CloseQuantumMachineResponse.CloseQuantumMachineResponseSuccess
    error: CloseQuantumMachineResponse.CloseQuantumMachineResponseError
    def __init__(self, success: _Optional[_Union[CloseQuantumMachineResponse.CloseQuantumMachineResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[CloseQuantumMachineResponse.CloseQuantumMachineResponseError, _Mapping]] = ...) -> None: ...

class CloseAllQuantumMachinesResponse(_message.Message):
    __slots__ = ("success", "error")
    class CloseAllQuantumMachinesResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class CloseAllQuantumMachinesResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: CloseAllQuantumMachinesResponse.CloseAllQuantumMachinesResponseSuccess
    error: CloseAllQuantumMachinesResponse.CloseAllQuantumMachinesResponseError
    def __init__(self, success: _Optional[_Union[CloseAllQuantumMachinesResponse.CloseAllQuantumMachinesResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[CloseAllQuantumMachinesResponse.CloseAllQuantumMachinesResponseError, _Mapping]] = ...) -> None: ...

class ClearAllJobResultsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ClearAllJobResultsResponse(_message.Message):
    __slots__ = ("success", "error")
    class ClearAllJobResultsResponseSuccess(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    class ClearAllJobResultsResponseError(_message.Message):
        __slots__ = ("details",)
        DETAILS_FIELD_NUMBER: _ClassVar[int]
        details: str
        def __init__(self, details: _Optional[str] = ...) -> None: ...
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    success: ClearAllJobResultsResponse.ClearAllJobResultsResponseSuccess
    error: ClearAllJobResultsResponse.ClearAllJobResultsResponseError
    def __init__(self, success: _Optional[_Union[ClearAllJobResultsResponse.ClearAllJobResultsResponseSuccess, _Mapping]] = ..., error: _Optional[_Union[ClearAllJobResultsResponse.ClearAllJobResultsResponseError, _Mapping]] = ...) -> None: ...



from typing import Type, TypeVar

from google.protobuf.any_pb2 import Any
from google.protobuf.message import Message

ProtoMessageT = TypeVar("ProtoMessageT", bound=Message)


def proto_to_any(message: Message) -> Any:
    """
    Packs a standard protobuf message into a protobuf Any message.

    Args:
        message (Message): A protobuf message instance.

    Returns:
        Any: The packed Any message.
    """
    any_pb_message = Any()
    any_pb_message.Pack(message)
    return any_pb_message


def any_to_proto(any_message: Any, message_cls: Type[ProtoMessageT]) -> ProtoMessageT:
    """
    Converts a protobuf Any message into a standard protobuf message.

    Args:
        any_message (Any): The protobuf Any message.
        message_cls (Type[ProtoMessageT]): The expected protobuf message class to unpack into.

    Returns:
        Message: The unpacked protobuf message.
    """
    message = message_cls()
    any_message.Unpack(message)
    return message



from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.utils.protobuf_utils import serialized_on_wire
from qm.type_hinting.config_types import OscillatorConfigType
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter


class OscillatorConverter(BaseDictToPbConverter[OscillatorConfigType, inc_qua_config_pb2.QuaConfig.Oscillator]):
    def convert(self, input_data: OscillatorConfigType) -> inc_qua_config_pb2.QuaConfig.Oscillator:
        return self.oscillator_to_pb(input_data)

    def oscillator_to_pb(self, data: OscillatorConfigType) -> inc_qua_config_pb2.QuaConfig.Oscillator:
        oscillator = inc_qua_config_pb2.QuaConfig.Oscillator()
        if "intermediate_frequency" in data:
            oscillator.intermediateFrequency.value = int(data["intermediate_frequency"])
            if self._capabilities.supports_double_frequency:
                oscillator.intermediateFrequencyDouble = float(data["intermediate_frequency"])

        if "mixer" in data:
            oscillator.mixer.CopyFrom(inc_qua_config_pb2.QuaConfig.MixerRef(mixer=data["mixer"]))
            oscillator.mixer.loFrequency = int(data.get("lo_frequency", 0))
            if self._capabilities.supports_double_frequency:
                oscillator.mixer.loFrequencyDouble = float(data.get("lo_frequency", 0.0))

        return oscillator

    def deconvert(self, output_data: inc_qua_config_pb2.QuaConfig.Oscillator) -> OscillatorConfigType:
        oscillator_config_data: OscillatorConfigType = {}
        if output_data.intermediateFrequencyDouble:
            freq = output_data.intermediateFrequencyDouble
            oscillator_config_data["intermediate_frequency"] = freq
        elif output_data.intermediateFrequency:
            freq = int(output_data.intermediateFrequency.value)
            oscillator_config_data["intermediate_frequency"] = freq
        if serialized_on_wire(output_data.mixer):
            if output_data.mixer.mixer:
                oscillator_config_data["mixer"] = output_data.mixer.mixer
            if output_data.mixer.loFrequencyDouble:
                lo_freq = output_data.mixer.loFrequencyDouble
                oscillator_config_data["lo_frequency"] = float(lo_freq)
            elif output_data.mixer.loFrequency:
                lo_freq = output_data.mixer.loFrequency
                oscillator_config_data["lo_frequency"] = int(lo_freq)
        return oscillator_config_data



"""
A module that provides functionality for creating function expressions, a specific type of AnyScalarExpression.
Currently, all function expressions are exclusively accessible to users through the broadcast object.
This approach is a newer and improved alternative to the functionality provided by LibFunctions, and it works in
much the same way. However, unlike LibFunctions, each function in this class has its own dedicated Proto message for
its description. This results in a significantly clearer API, as opposed to LibFunctions, where all functions share
the same Proto message, making it more challenging to understand the purpose of individual functions.
"""

from typing import List, Union

from qm._loc import _get_loc
from qm.type_hinting import NumberT
from qm.grpc.qm.pb import inc_qua_pb2
from qm.qua._expressions import Scalar, QuaArrayVariable, QuaFunctionOutput, create_qua_scalar_expression


def _standardize_args(
    *args: Union[Scalar[NumberT], QuaArrayVariable[NumberT]]
) -> List[inc_qua_pb2.QuaProgram.FunctionExpression.ScalarOrVectorArgument]:
    standardized_args = []

    for arg in args:
        if isinstance(arg, QuaArrayVariable):
            standardized_args.append(
                inc_qua_pb2.QuaProgram.FunctionExpression.ScalarOrVectorArgument(array=arg.unwrapped)
            )
        else:
            arg = create_qua_scalar_expression(arg)
            standardized_args.append(
                inc_qua_pb2.QuaProgram.FunctionExpression.ScalarOrVectorArgument(scalar=arg.unwrapped)
            )

    return standardized_args


def and_(*values: Union[Scalar[bool], QuaArrayVariable[bool]]) -> QuaFunctionOutput[bool]:
    """Performs a logical AND operation on the input values.

    Args:
        *values (boolean, QUA boolean, Qua array of type boolean): The input values to be combined using a
            logical AND operation. Each input can be a single boolean, a QUA boolean or a QUA array of booleans.
    """
    function_expression = inc_qua_pb2.QuaProgram.FunctionExpression(loc=_get_loc())
    getattr(function_expression, "and").CopyFrom(
        inc_qua_pb2.QuaProgram.FunctionExpression.AndFunction(values=_standardize_args(*values))
    )
    return QuaFunctionOutput(function_expression, bool)


def or_(*values: Union[Scalar[bool], QuaArrayVariable[bool]]) -> QuaFunctionOutput[bool]:
    """Performs a logical OR operation on the input values.

    Args:
        *values (boolean, QUA boolean, Qua array of type boolean): The input values to be combined using a
            logical OR operation. Each input can be a single boolean, a QUA boolean or a QUA array of booleans.
    """
    function_expression = inc_qua_pb2.QuaProgram.FunctionExpression(loc=_get_loc())
    getattr(function_expression, "or").CopyFrom(
        inc_qua_pb2.QuaProgram.FunctionExpression.OrFunction(values=_standardize_args(*values))
    )
    return QuaFunctionOutput(function_expression, bool)


def xor_(*values: Union[Scalar[bool], QuaArrayVariable[bool]]) -> QuaFunctionOutput[bool]:
    """Performs a logical XOR (exclusive OR) operation on the input values.

    Args:
        *values (boolean, QUA boolean, Qua array of type boolean): The input values to be combined using a
            logical XOR operation. Each input can be a single boolean, a QUA boolean, or a QUA array of booleans.
    """
    function_expression = inc_qua_pb2.QuaProgram.FunctionExpression(
        xor=inc_qua_pb2.QuaProgram.FunctionExpression.XorFunction(values=_standardize_args(*values)),
        loc=_get_loc(),
    )
    return QuaFunctionOutput(function_expression, bool)



import logging
from collections.abc import Iterable
from datetime import datetime, timezone
from typing import Any, Union, TypeVar, MutableMapping, MutableSequence

from google.protobuf.message import Message
from google.protobuf.timestamp_pb2 import Timestamp

from qm.exceptions import QmQuaException
from qm.grpc.qm.pb import general_messages_pb2

LOG_LEVEL_MAP = {
    general_messages_pb2.MessageLevel.Message_LEVEL_ERROR: logging.ERROR,
    general_messages_pb2.MessageLevel.Message_LEVEL_WARNING: logging.WARN,
    general_messages_pb2.MessageLevel.Message_LEVEL_INFO: logging.INFO,
}


Node = Union[Message, Iterable["Node"]]
PROTOBUF_SCALAR_TYPES = (int, float, bool, str, bytes)


def list_fields(node: Message) -> dict[str, Node]:
    output: dict[str, Node] = {}
    for field in node.DESCRIPTOR.fields:
        field_value = getattr(node, field.name)
        if isinstance(field_value, PROTOBUF_SCALAR_TYPES):
            continue
        if isinstance(field_value, Iterable):
            output[field.name] = field_value
        elif node.HasField(field.name):
            if isinstance(field_value, Message) and serialized_on_wire(field_value):
                output[field.name] = field_value

    return output


def which_one_of(message: Message, oneof_group: str) -> tuple[str, Union[Any, None]]:
    """Return the name of the field set in the oneof group, or None if none is set."""
    oneof_name = message.WhichOneof(oneof_group)
    if oneof_name is None:
        return "None", None
    return oneof_name, getattr(message, oneof_name)


K = TypeVar("K")
V = TypeVar("V")


def update_map(map_container: MutableMapping[K, V], data: dict[K, V]) -> None:
    """Update the map with new values."""
    for key, value in data.items():
        if value:
            if isinstance(value, PROTOBUF_SCALAR_TYPES):
                map_container[key] = value
            elif isinstance(value, Message):
                item = map_container[key]
                if not isinstance(item, Message):
                    raise QmQuaException(
                        f"not matching map value to dict value, expected {type(value)}, got {type(item)}"
                    )
                item.CopyFrom(value)
            else:
                raise QmQuaException(f"not supported protobuf type: {type(value)}")
        else:
            del map_container[key]


def proto_map_to_dict(map_container: MutableMapping[K, V]) -> dict[K, V]:
    return {k: v for k, v in map_container.items()}


def assign_map(map_container: MutableMapping[K, V], data: dict[K, V]) -> None:
    """Assign a mapping to a map field in the protobuf message."""
    if hasattr(map_container, "clear"):
        map_container.clear()
    else:
        for k in list(map_container.keys()):
            del map_container[k]
    update_map(map_container, data)


T = TypeVar("T")


def proto_repeated_to_list(repeated_container: MutableSequence[T]) -> list[T]:
    return [v for v in repeated_container]


def assign_repeated(repeated_container: MutableSequence[T], items: Iterable[T]) -> None:
    """Assign a sequence to a repeated field in the protobuf message."""
    if hasattr(repeated_container, "clear"):
        repeated_container.clear()
    else:
        # Fallback for older protobuf versions without clear()
        del repeated_container[:]
    repeated_container.extend(items)


def assign_to_proto(msg: Message, field_name: str, value: Any) -> None:
    value_field = getattr(msg, field_name)
    if isinstance(value_field, MutableMapping):
        assign_map(value_field, value)
    elif isinstance(value_field, MutableSequence):
        assign_repeated(value_field, value)
    elif isinstance(value_field, Message):
        value_field.CopyFrom(value)
    elif isinstance(value_field, PROTOBUF_SCALAR_TYPES):
        setattr(msg, field_name, value)
    else:
        raise QmQuaException(f"not supported protobuf type: {type(value_field)}")


def serialized_on_wire(message: Message) -> bool:
    """Check if message would be serialized (has non-default content)"""
    serialized = message.SerializeToString()
    return len(serialized) > 0


def timestamp_to_datetime(ts: Timestamp) -> datetime:
    # Convert protobuf Timestamp → Python datetime (UTC)
    return ts.ToDatetime().replace(tzinfo=timezone.utc)



from typing import Any, Optional, overload

import numpy as np

from qm._loc import _get_loc
from qm.exceptions import QmQuaException
from qm.type_hinting.general import Number
from qm.qua._scope_management._core_scopes import _PythonNativeScope
from qm.qua.extensions.qua_iterators.qua_iterators_base import IterableBase
from qm.qua.extensions.qua_iterators.qua_iterators_types import V, MetaDataType, NativeArrayType, NativeIteratorType


class NativeIterableBase(IterableBase[V]):
    def __init__(self, name: str, metadata: Optional[MetaDataType]):
        super().__init__(name, metadata)
        self._create_native_scope = True

    def __iter__(self) -> NativeIteratorType[V]:
        native_scope = _PythonNativeScope(_get_loc())
        with native_scope:
            self._add_to_current_scope()
            for i, v in enumerate(self.values):
                native_scope.set_current_iteration_number(i)
                yield v


class NativeIterableRange(NativeIterableBase[Any]):
    """
    Python-side range iterator.

    Use this helper for sweep dimensions that should stay in Python, for
    example when choosing configuration variants or element names around QUA
    code. No QUA variable is declared; the yielded value is a Python value.

    Use ``NativeIterableRange`` instead of a plain Python iterator when the
    Python-side axis should still participate in Sweep Program composition. The
    iterable name is preserved, so the axis can be combined inside
    [QuaProduct][qm.qua.extensions.qua_iterators.QuaProduct] or
    [QuaZip][qm.qua.extensions.qua_iterators.QuaZip], and its current value can
    be reflected in generated stream names by
    [declare_with_stream][qm.qua.declare_with_stream].

    ``NativeIterableRange(name, stop)`` behaves like ``range(stop)``.
    ``NativeIterableRange(name, start, stop, step)`` behaves like
    ``range(start, stop, step)`` for integers and like
    ``numpy.arange(start, stop, step)`` for floats. ``stop`` is exclusive.

    Example:
        ```python
        with program() as prog:
            for amp_scale in NativeIterableRange("amp_scale", 0.1, 1.0, 0.05):
                play("pulse" * amp(amp_scale), "element")
        ```

        This is equivalent to:

        ```python
        with program() as prog:
            for amp_scale in np.arange(0.1, 1.0, 0.05):
                play("pulse" * amp(amp_scale), "element")
        ```
    """

    @overload
    def __init__(self, name: str, stop: int, *, metadata: Optional[MetaDataType] = None):
        """
        support range(5)
        """
        ...

    @overload
    def __init__(
        self, name: str, start: Number, stop: Number, step: Number = 1, *, metadata: Optional[MetaDataType] = None
    ):
        ...

    def __init__(self, name: str, *args: Number, metadata: Optional[MetaDataType] = None):  # type: ignore[misc]
        super().__init__(name, metadata)
        self._start: Number
        self._stop: Number
        self._step: Number = 1

        if len(args) == 1:
            self._stop = args[0]
            self._start = 0
        elif len(args) == 2:
            self._start, self._stop = args
        elif len(args) == 3:
            self._start, self._stop, self._step = args
        else:
            raise QmQuaException(f"not supported args len = {len(args)}")

    @property
    def values(self) -> NativeArrayType:
        """
        Return the Python values represented by this iterable.
        """
        return np.arange(self._start, self._stop, self._step).tolist()


class NativeIterable(NativeIterableBase[Any]):
    """
    Python-side iterable over explicit values.

    Use this helper for loops that choose Python values such as element names,
    labels, or configuration variants. No QUA variable is declared; the yielded
    value is a Python value.

    Use ``NativeIterable`` instead of a plain Python iterator when the
    Python-side axis should still participate in Sweep Program composition. The
    iterable name is preserved, so the axis can be combined inside
    [QuaProduct][qm.qua.extensions.qua_iterators.QuaProduct] or
    [QuaZip][qm.qua.extensions.qua_iterators.QuaZip], and its current value can
    be reflected in generated stream names by
    [declare_with_stream][qm.qua.declare_with_stream].

    Example:
        ```python
        with program() as prog:
            for element in NativeIterable("element", ["q1", "q2"]):
                play("x180", element)
        ```

        This is equivalent to:

        ```python
        with program() as prog:
            for element in ["q1", "q2"]:
                play("x180", element)
        ```
    """

    def __init__(self, name: str, array: NativeArrayType, *, metadata: Optional[MetaDataType] = None):
        super().__init__(name, metadata)
        self._array: NativeArrayType = array

    @property
    def values(self) -> NativeArrayType:
        """
        Return the Python values represented by this iterable.
        """
        return self._array



from typing import Union, cast
from collections.abc import Iterable, Collection

from qm._loc import _get_loc
from qm.type_hinting import NumberT
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.qua._dsl._type_hints import OneOrMore
from qm.qua._dsl.variable_handling import declare
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.utils.protobuf_utils import which_one_of, serialized_on_wire
from qm.qua._scope_management._core_scopes import _ProgramScope, _ResultAnalysisScope
from qm.utils import collection_has_type_int, collection_has_type_bool, collection_has_type_float
from qm.qua._dsl.stream_processing.direct_stream_processing_interface import DirectStreamSourceInterface
from qm.qua._expressions import (
    Scalar,
    Vector,
    QuaScalar,
    QuaVariable,
    QuaArrayVariable,
    QuaNumericExpression,
    fixed,
    literal_bool,
    to_scalar_pb_expression,
    create_qua_scalar_expression,
)
from qm.qua._scope_management.scopes import (
    _IfScope,
    _ForScope,
    _CaseScope,
    _ElifScope,
    _ElseScope,
    _SwitchScope,
    _ForEachScope,
    _CaseDefaultScope,
    _StrictTimingScope,
    _PortConditionScope,
)


def program() -> _ProgramScope:
    """Create a QUA program.

    Used within a context manager, the program is defined in the code block
    below ``with`` statement.

    Statements in the code block below are played as soon as possible, meaning that an instruction
    will be played immediately unless it is dependent on a previous instruction.
    Additionally, commands output to the same elements will be played sequentially,
    and to different elements will be played in parallel.
    An exception is that pulses will be implicitly aligned at the end of each [`for_`][qm.qua.for_] loop iteration.

    The generated ``program_name`` object is used as an input to the execution or compilation functions.

    Example:
        ```python
        with program() as program_name:
            play('pulse1', 'element1')
            wait('element1')

        qmm = QuantumMachinesManager(...)
        qm = qmm.open_qm(...)
        qm.execute(program_name)
        ```
    """
    return _ProgramScope()


def switch_(expression: QuaScalar[NumberT], unsafe: bool = False) -> _SwitchScope[NumberT]:
    """Part of the switch-case flow control statement in QUA.

    To be used with a context manager.

    The code block inside should be composed of only ``case_()`` and ``default_()``
    statements, and there should be at least one of them.

    The expression given in the ``switch_()`` statement will be evaluated and compared
    to each of the values in the ``case_()`` statements. The QUA code block following
    the ``case_()`` statement which evaluated to true will be executed. If none of the
    statements evaluated to true, the QUA code block following the ``default_()``
    statement (if given) will be executed.

    Args:
        expression: An expression to evaluate
        unsafe: If set to True, then switch-case would be more efficient
            and would produce fewer gaps. However, if an input which does
            not match a case is given, unexpected behavior will occur.
            Cannot be used with the ``default_()`` statement. Default is
            false, use with care.

    Example:
        ```python
        x=declare(int)
        with switch_(x):
            with case_(1):
                play('first_pulse', 'element')
            with case_(2):
                play('second_pulse', 'element')
            with case_(3):
                play('third_pulse', 'element')
            with default_():
                play('other_pulse', 'element')
        ```
    """
    return _SwitchScope(expression, unsafe, loc=_get_loc())


def case_(case_exp: Scalar[NumberT]) -> _CaseScope[NumberT]:
    """Part of the switch-case flow control statement in QUA.

    To be used with a context manager.

    Must be inside a ``switch_()`` statement.

    The expression given in the ``switch_()`` statement will be evaluated and compared
    to each of the values in the ``case_()`` statements. The QUA code block following
    the ``case_()`` statement which evaluated to true will be executed. If none of the
    statements evaluated to true, the QUA code block following the ``default_()``
    statement (if given) will be executed.

    Args:
        case_exp: A value (or expression) to compare to the expression
            in the ``switch_()`` statement

    Example:
        ```python
        x=declare(int)
        with switch_(x):
            with case_(1):
                play('first_pulse', 'element')
            with case_(2):
                play('second_pulse', 'element')
            with case_(3):
                play('third_pulse', 'element')
            with default_():
                play('other_pulse', 'element')
        ```
    """
    switch = scopes_manager.current_scope
    if not isinstance(switch, _SwitchScope):
        raise QmQuaException("Expecting switch scope. 'case_' can be used only under switch scope.")

    if switch.default:
        raise QmQuaException("All 'case' statements must precede the 'default' statement.")

    return _CaseScope(switch, create_qua_scalar_expression(case_exp))


def default_() -> _CaseDefaultScope:
    """Part of the switch-case flow control statement in QUA.

    To be used with a context manager.

    Must be inside a ``switch_()`` statement, and there can only be one ``default_()``
    statement.

    The expression given in the ``switch_()`` statement will be evaluated and compared
    to each of the values in the ``case_()`` statements. The QUA code block following
    the ``case_()`` statement which evaluated to true will be executed. If none of the
    statements evaluated to true, the QUA code block following the ``default_()``
    statement (if given) will be executed.

    Example:
        ```python
        x=declare(int)
        with switch_(x):
            with case_(1):
                play('first_pulse', 'element')
            with case_(2):
                play('second_pulse', 'element')
            with case_(3):
                play('third_pulse', 'element')
            with default_():
                play('other_pulse', 'element')
        ```
    """
    switch = scopes_manager.current_scope
    if not isinstance(switch, _SwitchScope):
        raise QmQuaException("Expecting switch scope. 'default_' can be used only under switch scope.")

    if not switch.cases:
        raise QmQuaException("must specify at least one case before 'default'.")

    if switch.default:
        raise QmQuaException("only a single 'default' statement can follow a 'switch' statement")

    return _CaseDefaultScope(switch)


def if_(expression: Scalar[bool], unsafe: bool = False) -> _IfScope:
    """If flow control statement in QUA.

    To be used with a context manager.

    The QUA code block following the statement will be
    executed only if the expression given evaluates to true.

    Args:
        expression: A boolean expression to evaluate

    Example:
        ```python
        x=declare(int)
        with if_(x>0):
            play('pulse', 'element')
        ```
    """
    return _IfScope(expression, unsafe, loc=_get_loc())


def elif_(expression: Scalar[bool]) -> _ElifScope:
    """Else-If flow control statement in QUA.

    To be used with a context manager.

    Must appear after an ``if_()`` statement.

    The QUA code block following the statement will be executed only if the expressions
    in the preceding ``if_()`` and ``elif_()`` statements evaluates to false and if the
    expression given in this ``elif_()`` evaluates to true.

    Args:
        expression: A boolean expression to evaluate

    Example:
        ```python
        x=declare(int)
        with if_(x>2):
            play('pulse', 'element')
        with elif_(x>-2):
            play('other_pulse', 'element')
        with else_():
            play('third_pulse', 'element')
        ```
    """
    scope = scopes_manager.current_scope
    if not scope.statements:
        raise QmQuaException(
            "'elif' statement must directly follow 'if' statement - Please make sure it is aligned with the corresponding if statement."
        )
    _, if_statement = which_one_of(scope.statements[-1], "statement_oneof")
    if not isinstance(if_statement, inc_qua_pb2.QuaProgram.IfStatement):
        raise QmQuaException(
            "'elif' statement must directly follow 'if' statement - Please make sure it is aligned with the corresponding if statement."
        )

    if serialized_on_wire(getattr(if_statement, "else")):
        raise QmQuaException("'elif' must come before 'else' statement")

    return _ElifScope(condition=expression, if_statement=if_statement, loc=if_statement.loc)


def else_() -> _ElseScope:
    """Else flow control statement in QUA.

    To be used with a context manager.

    Must appear after an ``if_()`` statement.

    The QUA code block following the statement will be executed only if the expressions
    in the preceding ``if_()`` and ``elif_()`` statements evaluates to false.

    Example:
        ```python
        x=declare(int)
        with if_(x>0):
            play('pulse', 'element')
        with else_():
            play('other_pulse', 'element')
        ```
    """
    scope = scopes_manager.current_scope
    if not scope.statements:
        raise QmQuaException(
            "'else' statement must directly follow 'if' statement - "
            "Please make sure it is aligned with the corresponding if statement."
        )
    _, if_statement = which_one_of(scope.statements[-1], "statement_oneof")
    if not isinstance(if_statement, inc_qua_pb2.QuaProgram.IfStatement):
        raise QmQuaException(
            "'else' statement must directly follow 'if' statement - "
            "Please make sure it is aligned with the corresponding if statement."
        )

    if serialized_on_wire(getattr(if_statement, "else")):
        raise QmQuaException("only a single 'else' statement can follow an 'if' statement")

    return _ElseScope(if_statement)


def while_(cond: QuaScalar[bool]) -> _ForScope:
    """While loop flow control statement in QUA.

    To be used with a context manager.

    Args:
        cond (QUA expression): an expression which evaluates to a
            boolean variable, determines if to continue to next loop
            iteration

    Example:
        ```python
        x = declare(int)
        assign(x, 0)
        with while_(x<=30):
            play('pulse', 'element')
            assign(x, x+1)
        ```
    """
    return _ForScope(None, to_scalar_pb_expression(cond), None, loc=_get_loc())


def for_(
    var: Union[QuaVariable[NumberT], DirectStreamSourceInterface[NumberT]],
    init: Scalar[NumberT],
    cond: QuaScalar[bool],
    update: QuaScalar[NumberT],
) -> _ForScope:
    """For loop flow control statement in QUA.

    To be used with a context manager.

    Args:
        var (QUA variable): QUA variable used as iteration variable
        init (QUA expression): an expression which sets the initial
            value of the iteration variable
        cond (QUA expression): an expression which evaluates to a
            boolean variable, determines if to continue to next loop
            iteration
        update (QUA expression): an expression to assign to ``var`` with
            each loop iteration

    Example:
        ```python
        x = declare(fixed)
        with for_(var=x, init=0, cond=x<=1, update=x+0.1):
            play('pulse', 'element')
        ```
    """
    init_statement = inc_qua_pb2.QuaProgram.AssignmentStatement(
        target=var.assignment_statement,
        expression=to_scalar_pb_expression(init),
        loc=_get_loc(),
    )
    update_statement = inc_qua_pb2.QuaProgram.AssignmentStatement(
        target=var.assignment_statement, expression=to_scalar_pb_expression(update), loc=_get_loc()
    )
    cond_expression = to_scalar_pb_expression(cond)

    for_scope = _ForScope(init_statement, cond_expression, update_statement, loc=_get_loc())

    # append save statement if var is DirectStreamSourceInterface
    if isinstance(var, DirectStreamSourceInterface):
        for_scope.append_statement(var.get_save_statement())

    return for_scope


def for_each_(
    var: OneOrMore[Union[QuaVariable[NumberT], DirectStreamSourceInterface[NumberT]]],
    values: OneOrMore[Vector[NumberT]],
) -> _ForEachScope:
    """Flow control: Iterate over array elements in QUA.

    It is possible to either loop over one variable, or over a tuple of variables,
    similar to the `zip` style iteration in python.

    To be used with a context manager.

    Args:
        var (Union[QUA variable, tuple of QUA variables]): The iteration
            variable
        values (Union[list of literals, tuple of lists of literals, QUA array, tuple of QUA arrays]):
            A list of values to iterate over or a QUA array.

    Example:
        ```python
        x=declare(fixed)
        y=declare(fixed)
        with for_each_(x, [0.1, 0.4, 0.6]):
            play('pulse' * amp(x), 'element')
        with for_each_((x, y), ([0.1, 0.4, 0.6], [0.3, -0.2, 0.1])):
            play('pulse1' * amp(x), 'element')
            play('pulse2' * amp(y), 'element')
        ```

    Warning:

        This behavior is not exactly consistent with python `zip`.
        Instead of sending a list of tuple as values, the function expects a tuple of
        lists.
        The first list containing the values for the first variable, and so on.
    """
    if not isinstance(var, Iterable):
        var = (var,)

    for i, v in enumerate(var):
        if not isinstance(v, QuaVariable):
            raise QmQuaException(f"for_each_ var {i} must be a variable")

    qua_expression_cond = isinstance(values, QuaNumericExpression)
    not_iterable_cond = not isinstance(values, (Iterable, QuaArrayVariable))
    tuple_of_non_iterables_cond = not isinstance(values[0], (Iterable, QuaArrayVariable))
    if qua_expression_cond or not_iterable_cond or tuple_of_non_iterables_cond:
        values = (cast(QuaArrayVariable[NumberT], values),)
    values = cast(tuple[QuaArrayVariable[NumberT], ...], values)

    if isinstance(values, Collection) and len(values) < 1:
        raise QmQuaException("values cannot be empty")

    if len(var) != len(values):
        raise QmQuaException("number of variables does not match number of array values")

    arrays: list[Union[QuaArrayVariable[bool], QuaArrayVariable[int], QuaArrayVariable[float]]] = []
    for value in values:
        if isinstance(value, QuaArrayVariable):
            arrays.append(value)
        elif isinstance(value, Collection):
            has_bool = collection_has_type_bool(value)
            has_int = collection_has_type_int(value)
            has_float = collection_has_type_float(value)

            if has_bool:
                if has_int or has_float:
                    raise QmQuaException("values can not contain both bool and number values")
                # Only booleans
                arrays.append(declare(bool, value=[bool(x) for x in value]))
            else:
                if has_float:
                    # All will be considered as fixed
                    arrays.append(declare(fixed, value=[float(x) for x in value]))
                else:
                    # Only ints
                    arrays.append(declare(int, value=[int(x) for x in value]))
        else:
            raise QmQuaException("value is not a QUA array neither iterable")

    unwrapped_vars = [v.unwrapped.variable for v in var]
    unwrapped_arrays = [a.unwrapped for a in arrays]

    iterators = [(unwrapped_vars[i], ar) for i, ar in enumerate(unwrapped_arrays)]
    for_each_scope = _ForEachScope(iterators, loc=_get_loc())
    # append save statements for DirectStreamSourceInterface variables
    for v in var:
        if isinstance(v, DirectStreamSourceInterface):
            for_each_scope.append_statement(v.get_save_statement())

    return for_each_scope


def infinite_loop_() -> _ForScope:
    """Infinite loop flow control statement in QUA.

    To be used with a context manager.

    Optimized for zero latency between iterations,
    provided that no more than a single element appears in the loop.

    Note:
        In case multiple elements need to be used in an infinite loop, it is possible to add several loops
        in parallel (see example).
        Two infinite loops cannot share an element nor can they share variables.

    Example:
        ```python
        with infinite_loop_():
            play('pulse1', 'element1')
        with infinite_loop_():
            play('pulse2', 'element2')
        ```
    """
    return _ForScope(None, literal_bool(True), None, loc=_get_loc())


def port_condition(condition: Scalar[bool]) -> _PortConditionScope:
    """
    A context manager for a faster conditional play mechanism. Will operate on all the elements inside the context manager.
    Note that elements sharing a port with an element inside the context manager cannot be played in parallel to the context manager.

    -- Available for MW-FEM Only --

    Args:
        condition (A logical expression to evaluate): Will play the operation only if the condition is true.
            The play command will take the same amount of time regardless of the condition (If false, would wait instead).

    Example:
        ```python
        with port_condition(x > 0):
            play('pulse', 'element')
        ```
    """
    return _PortConditionScope(expression=to_scalar_pb_expression(condition))


def stream_processing() -> _ResultAnalysisScope:
    """A context manager for the creation of [Stream processing pipelines](../../Guides/stream_proc.md#overview)

    Each pipeline defines an analysis process that is applied to every stream item.
    A pipeline must be terminated with a save/save_all terminal, and then can be retrieved with
    [QmJob.result_handles][qm.jobs.running_qm_job.RunningQmJob.result_handles].

    There are two save options: ``save_all`` will save every stream item, ``save`` will save only last item.

    A pipeline can be assigned to python variable, and then reused on other pipelines. It is ensured that the
    common part of the pipeline is processed only once.

    ??? example "Creating a result analysis object"
        ```python
        with stream_processing():
            a.save("tag")
            a.save_all("another tag")
        ```

    ??? example "Retrieving saved result"
        ```python
        QmJob.result_handles.get("tag")
        ```
    """
    scope = scopes_manager.program_scope
    return scope.result_analysis_scope


def strict_timing_() -> _StrictTimingScope:
    """Any QUA command written within the strict timing block will be required to play without gaps.

    See [the documentation](../../Guides/timing_in_qua.md#strict-timing) for further information and examples.

    To be used with a context manager.

    -- Available from QOP 2.0 --
    """
    return _StrictTimingScope(loc=_get_loc())




# Wrapper to enforce deterministic serialization
def deterministic_serialize_wrapper(orig_serialize_func):
    def wrapped(msg):
        return orig_serialize_func(msg, deterministic=True)
    return wrapped

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from qm.grpc.octave.v1 import api_pb2 as octave_dot_v1_dot_api__pb2

GRPC_GENERATED_VERSION = '1.68.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in octave/v1/api_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class OctaveServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SubscribeState = channel.unary_stream(
                '/quantummachines.octave.api.v1.OctaveService/SubscribeState',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.SubscribeStateRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.SubscribeStateResponse.FromString,
                _registered_method=True)
        self.GetVersion = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/GetVersion',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.GetVersionRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.GetVersionResponse.FromString,
                _registered_method=True)
        self.Control = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Control',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.ControlRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.ControlResponse.FromString,
                _registered_method=True)
        self.Update = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Update',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.UpdateRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.UpdateResponse.FromString,
                _registered_method=True)
        self.Aquire = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Aquire',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.AquireRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.AquireResponse.FromString,
                _registered_method=True)
        self.Identify = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Identify',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.IdentifyRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.IdentifyResponse.FromString,
                _registered_method=True)
        self.Monitor = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Monitor',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.MonitorRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.MonitorResponse.FromString,
                _registered_method=True)
        self.Explore = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Explore',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.ExploreRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.ExploreResponse.FromString,
                _registered_method=True)
        self.Save = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Save',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.SaveRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.SaveResponse.FromString,
                _registered_method=True)
        self.Recall = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Recall',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.RecallRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.RecallResponse.FromString,
                _registered_method=True)
        self.List = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/List',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.ListRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.ListResponse.FromString,
                _registered_method=True)
        self.Reset = channel.unary_unary(
                '/quantummachines.octave.api.v1.OctaveService/Reset',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.ResetRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.ResetResponse.FromString,
                _registered_method=True)
        self.Health = channel.stream_stream(
                '/quantummachines.octave.api.v1.OctaveService/Health',
                request_serializer=deterministic_serialize_wrapper(octave_dot_v1_dot_api__pb2.HealthRequest.SerializeToString),
                response_deserializer=octave_dot_v1_dot_api__pb2.HealthResponse.FromString,
                _registered_method=True)


class OctaveServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SubscribeState(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Control(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Update(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Aquire(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Identify(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Monitor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Explore(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Save(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Recall(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def List(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Reset(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Health(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_OctaveServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SubscribeState': grpc.unary_stream_rpc_method_handler(
                    servicer.SubscribeState,
                    request_deserializer=octave_dot_v1_dot_api__pb2.SubscribeStateRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.SubscribeStateResponse.SerializeToString,
            ),
            'GetVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVersion,
                    request_deserializer=octave_dot_v1_dot_api__pb2.GetVersionRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.GetVersionResponse.SerializeToString,
            ),
            'Control': grpc.unary_unary_rpc_method_handler(
                    servicer.Control,
                    request_deserializer=octave_dot_v1_dot_api__pb2.ControlRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.ControlResponse.SerializeToString,
            ),
            'Update': grpc.unary_unary_rpc_method_handler(
                    servicer.Update,
                    request_deserializer=octave_dot_v1_dot_api__pb2.UpdateRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.UpdateResponse.SerializeToString,
            ),
            'Aquire': grpc.unary_unary_rpc_method_handler(
                    servicer.Aquire,
                    request_deserializer=octave_dot_v1_dot_api__pb2.AquireRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.AquireResponse.SerializeToString,
            ),
            'Identify': grpc.unary_unary_rpc_method_handler(
                    servicer.Identify,
                    request_deserializer=octave_dot_v1_dot_api__pb2.IdentifyRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.IdentifyResponse.SerializeToString,
            ),
            'Monitor': grpc.unary_unary_rpc_method_handler(
                    servicer.Monitor,
                    request_deserializer=octave_dot_v1_dot_api__pb2.MonitorRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.MonitorResponse.SerializeToString,
            ),
            'Explore': grpc.unary_unary_rpc_method_handler(
                    servicer.Explore,
                    request_deserializer=octave_dot_v1_dot_api__pb2.ExploreRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.ExploreResponse.SerializeToString,
            ),
            'Save': grpc.unary_unary_rpc_method_handler(
                    servicer.Save,
                    request_deserializer=octave_dot_v1_dot_api__pb2.SaveRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.SaveResponse.SerializeToString,
            ),
            'Recall': grpc.unary_unary_rpc_method_handler(
                    servicer.Recall,
                    request_deserializer=octave_dot_v1_dot_api__pb2.RecallRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.RecallResponse.SerializeToString,
            ),
            'List': grpc.unary_unary_rpc_method_handler(
                    servicer.List,
                    request_deserializer=octave_dot_v1_dot_api__pb2.ListRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.ListResponse.SerializeToString,
            ),
            'Reset': grpc.unary_unary_rpc_method_handler(
                    servicer.Reset,
                    request_deserializer=octave_dot_v1_dot_api__pb2.ResetRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.ResetResponse.SerializeToString,
            ),
            'Health': grpc.stream_stream_rpc_method_handler(
                    servicer.Health,
                    request_deserializer=octave_dot_v1_dot_api__pb2.HealthRequest.FromString,
                    response_serializer=octave_dot_v1_dot_api__pb2.HealthResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'quantummachines.octave.api.v1.OctaveService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('quantummachines.octave.api.v1.OctaveService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class OctaveService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SubscribeState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/SubscribeState',
            octave_dot_v1_dot_api__pb2.SubscribeStateRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.SubscribeStateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/GetVersion',
            octave_dot_v1_dot_api__pb2.GetVersionRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.GetVersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Control(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Control',
            octave_dot_v1_dot_api__pb2.ControlRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.ControlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Update(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Update',
            octave_dot_v1_dot_api__pb2.UpdateRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.UpdateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Aquire(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Aquire',
            octave_dot_v1_dot_api__pb2.AquireRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.AquireResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Identify(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Identify',
            octave_dot_v1_dot_api__pb2.IdentifyRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.IdentifyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Monitor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Monitor',
            octave_dot_v1_dot_api__pb2.MonitorRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.MonitorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Explore(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Explore',
            octave_dot_v1_dot_api__pb2.ExploreRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.ExploreResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Save(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Save',
            octave_dot_v1_dot_api__pb2.SaveRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.SaveResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Recall(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Recall',
            octave_dot_v1_dot_api__pb2.RecallRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.RecallResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def List(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/List',
            octave_dot_v1_dot_api__pb2.ListRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.ListResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Reset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Reset',
            octave_dot_v1_dot_api__pb2.ResetRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.ResetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Health(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/quantummachines.octave.api.v1.OctaveService/Health',
            octave_dot_v1_dot_api__pb2.HealthRequest.SerializeToString,
            octave_dot_v1_dot_api__pb2.HealthResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from typing import TYPE_CHECKING, Any, Union, Literal, Mapping, Callable, Optional

from qm.exceptions import QmQuaException
from qm.grpc.qm.pb.inc_qua_pb2 import QuaProgram
from qm.utils.protobuf_utils import Node, which_one_of
from qm.serialization.qua_node_visitor import QuaNodeVisitor

if TYPE_CHECKING:
    from qm.serialization.qua_serializing_visitor import QuaSerializingVisitor


class ExpressionSerializingVisitor(QuaNodeVisitor):
    def __init__(self, visitor: Optional["QuaSerializingVisitor"]) -> None:
        self._out = ""
        self._visitor = visitor
        super().__init__()

    @property
    def _node_to_visit(self) -> Mapping[type, Callable[[Any], None]]:
        return {
            QuaProgram.LibFunctionExpression: self.visit_lib_function_expression,
            QuaProgram.LibFunctionExpression.Argument: self.visit_lib_function_argument,
            QuaProgram.VarRefExpression: self.visit_var_reference,
            QuaProgram.ArrayVarRefExpression: self.visit_array_var_reference,
            QuaProgram.ArrayCellRefExpression: self.visit_array_cell_reference,
            QuaProgram.ArrayLengthExpression: self.visit_array_length,
            QuaProgram.LiteralExpression: self.visit_literal,
            QuaProgram.BroadcastExpression: self.visit_broadcast_expression,
            QuaProgram.FunctionExpression: self.visit_function_expression,
            QuaProgram.FunctionExpression.ScalarOrVectorArgument: self.visit_function_scalar_or_vector_argument,
            QuaProgram.AnalogMeasureProcess.DemodIntegration: self.visit_analog_measurement_demod_integration,
            QuaProgram.AnalogMeasureProcess.BareIntegration: self.visit_analog_measure_bare_integration,
            QuaProgram.AnalogMeasureProcess.DualDemodIntegration: self.visit_analog_measurement_dual_demod_integration,
            QuaProgram.AnalogMeasureProcess.RawTimeTagging: self.visit_analog_measurement_raw_time_tagging,
            QuaProgram.AnalogMeasureProcess.HighResTimeTagging: self.visit_analog_measurement_high_res_time_tagging,
            QuaProgram.DigitalMeasureProcess.Counting: self.visit_digital_measurement_counting,
            QuaProgram.DigitalMeasureProcess.RawTimeTagging: self.visit_digital_measurement_raw_time_tagging,
            QuaProgram.BinaryExpression: self.visit_binary_expression,
            QuaProgram.GlobalVarRefExpression: self.visit_global_var_reference,
        }

    def visit_lib_function_expression(self, node: QuaProgram.LibFunctionExpression) -> None:
        args = [ExpressionSerializingVisitor(self._visitor).serialize(arg) for arg in node.arguments]
        if node.libraryName == "random":
            seed = args.pop(0)
            self._out = f"Random({seed}).{node.functionName}({', '.join(args)})"
        else:
            library_name = {
                "util": "Util",
                "math": "Math",
                "cast": "Cast",
            }.get(node.libraryName, None)
            function_name = {
                "cond": "cond",
                "unsafe_cast_fixed": "unsafe_cast_fixed",
                "unsafe_cast_bool": "unsafe_cast_bool",
                "unsafe_cast_int": "unsafe_cast_int",
                "to_int": "to_int",
                "to_bool": "to_bool",
                "to_fixed": "to_fixed",
                "mul_fixed_by_int": "mul_fixed_by_int",
                "mul_int_by_fixed": "mul_int_by_fixed",
                "log": "log",
                "pow": "pow",
                "div": "div",
                "exp": "exp",
                "pow2": "pow2",
                "ln": "ln",
                "log2": "log2",
                "log10": "log10",
                "sqrt": "sqrt",
                "inv_sqrt": "inv_sqrt",
                "inv": "inv",
                "MSB": "msb",
                "elu": "elu",
                "aelu": "aelu",
                "selu": "selu",
                "relu": "relu",
                "plrelu": "plrelu",
                "lrelu": "lrelu",
                "sin2pi": "sin2pi",
                "cos2pi": "cos2pi",
                "atan_2pi": "atan_2pi",
                "atan2_2pi": "atan2_2pi",
                "abs": "abs",
                "sin": "sin",
                "cos": "cos",
                "atan": "atan",
                "atan2": "atan2",
                "sum": "sum",
                "max": "max",
                "min": "min",
                "argmax": "argmax",
                "argmin": "argmin",
                "dot": "dot",
            }.get(node.functionName, None)
            if library_name is None:
                raise Exception(f"Unsupported library name {node.libraryName}")
            if function_name is None:
                raise Exception(f"Unsupported function name {node.functionName}")

            self._out = f"{library_name}.{function_name}({','.join(args)})"

    def visit_lib_function_argument(self, node: QuaProgram.LibFunctionExpression.Argument) -> None:
        _, value = which_one_of(node, "argument_oneof")
        if value is not None:
            self._out = ExpressionSerializingVisitor(self._visitor).serialize(value)

    def visit_var_reference(self, node: QuaProgram.VarRefExpression) -> None:
        var_ref = which_one_of(node, "var_oneof")[1]
        self._out = node.name if isinstance(var_ref, str) else f"IO{var_ref}"

    def visit_array_var_reference(self, node: QuaProgram.ArrayVarRefExpression) -> None:
        if self._visitor:
            for struct in self._visitor.structs:
                if node.name in struct.variables():
                    self._out = f"{struct.variable_name}. {node.name}"

        if self._out == "":
            self._out = node.name

    def visit_array_cell_reference(self, node: QuaProgram.ArrayCellRefExpression) -> None:
        var = ExpressionSerializingVisitor(self._visitor).serialize(node.arrayVar)
        index = ExpressionSerializingVisitor(self._visitor).serialize(node.index)
        self._out = f"{var}[{index}]"

    def visit_array_length(self, node: QuaProgram.ArrayLengthExpression) -> None:
        var = ExpressionSerializingVisitor(self._visitor).serialize(node.array)
        self._out = f"{var}.length()"

    def visit_literal(self, node: QuaProgram.LiteralExpression) -> None:
        self._out = node.value

    def visit_broadcast_expression(self, node: QuaProgram.BroadcastExpression) -> None:
        value = node.value
        serialized_value = ExpressionSerializingVisitor(self._visitor).serialize(value)
        self._out = f"broadcast. {serialized_value}"

    def visit_function_expression(self, node: QuaProgram.FunctionExpression) -> None:
        function_name, function_expression = which_one_of(node, "function_oneof")
        # During proto class generation, function names that are reserved keywords are suffixed with an underscore.
        # Conversely, functions already suffixed with an underscore (but are not reserved keywords) are saved without
        # the underscore.  For example, 'xor_' (not a reserved keyword) will be saved as 'xor'.
        if not function_name.endswith("_"):
            function_name += "_"

        # For mypy
        if not isinstance(
            function_expression,
            (
                QuaProgram.FunctionExpression.AndFunction,
                QuaProgram.FunctionExpression.OrFunction,
                QuaProgram.FunctionExpression.XorFunction,
            ),
        ):
            raise QmQuaException(f"Got non expected type for function_oneof, got {type(function_expression)}")

        function_arguments = [
            ExpressionSerializingVisitor(self._visitor).serialize(arg) for arg in function_expression.values
        ]

        self._out = f"{function_name}({', '.join(function_arguments)})"

    def visit_function_scalar_or_vector_argument(
        self, node: QuaProgram.FunctionExpression.ScalarOrVectorArgument
    ) -> None:
        _, value = which_one_of(node, "argument_oneof")
        if value is not None:
            self._out = ExpressionSerializingVisitor(self._visitor).serialize(value)

    def visit_analog_measurement_demod_integration(
        self, node: QuaProgram.AnalogMeasureProcess.DemodIntegration
    ) -> None:
        self._common_integration(node, "demod")

    def visit_analog_measure_bare_integration(self, node: QuaProgram.AnalogMeasureProcess.BareIntegration) -> None:
        self._common_integration(node, "integration")

    def _common_integration(
        self,
        node: Union[QuaProgram.AnalogMeasureProcess.BareIntegration, QuaProgram.AnalogMeasureProcess.DemodIntegration],
        integration_type: Literal["integration", "demod"],
    ) -> None:
        name = node.integration.name
        output = node.elementOutput
        target_name, target_value = which_one_of(node.target, "processTarget")

        if isinstance(target_value, QuaProgram.AnalogProcessTarget.ScalarProcessTarget):
            target = ExpressionSerializingVisitor(self._visitor).serialize(target_value)
            self._out = f'{integration_type}.full("{name}", {target}, "{output}")'
        elif isinstance(target_value, QuaProgram.AnalogProcessTarget.VectorProcessTarget):
            target = ExpressionSerializingVisitor(self._visitor).serialize(target_value.array)

            time_name, time_value = which_one_of(target_value.timeDivision, "timeDivision")
            if isinstance(time_value, QuaProgram.AnalogTimeDivision.Sliced):
                self._out = f'{integration_type}.sliced("{name}", {target}, {time_value.samplesPerChunk}, "{output}")'
            elif isinstance(time_value, QuaProgram.AnalogTimeDivision.Accumulated):
                self._out = (
                    f'{integration_type}.accumulated("{name}", {target}, {time_value.samplesPerChunk}, "{output}")'
                )
            elif isinstance(time_value, QuaProgram.AnalogTimeDivision.MovingWindow):
                self._out = f'{integration_type}.moving_window("{name}", {target}, {time_value.samplesPerChunk}, {time_value.chunksPerWindow}, "{output}")'
            else:
                raise Exception(f"Unsupported analog process target {target_name}")
        else:
            raise Exception(f"Unsupported analog process target {target_name}")

    def _common_dual_measurement(
        self,
        node: Union[
            QuaProgram.AnalogMeasureProcess.DualDemodIntegration,
            QuaProgram.AnalogMeasureProcess.DualBareIntegration,
        ],
        demod_name: str,
    ) -> None:
        name1 = node.integration1.name
        name2 = node.integration2.name
        output1 = node.elementOutput1
        output2 = node.elementOutput2
        target_name, target_value = which_one_of(node.target, "processTarget")
        if output1 == "out1" and output2 == "out2":
            common_args = f'"{name1}", "{name2}"'
        elif output1 == "out2" and output2 == "out1":
            common_args = f'"{name2}", "{name1}"'
        else:
            common_args = f'"{name1}", "{output1}", "{name2}", "{output2}"'
        if isinstance(target_value, QuaProgram.AnalogProcessTarget.ScalarProcessTarget):
            target = ExpressionSerializingVisitor(self._visitor).serialize(target_value)
            self._out = f"{demod_name}.full({common_args}, {target})"
        elif isinstance(target_value, QuaProgram.AnalogProcessTarget.VectorProcessTarget):
            target = ExpressionSerializingVisitor(self._visitor).serialize(target_value.array)

            time_name, time_value = which_one_of(target_value.timeDivision, "timeDivision")
            if isinstance(time_value, QuaProgram.AnalogTimeDivision.Sliced):
                self._out = f"{demod_name}.sliced({common_args}, {time_value.samplesPerChunk}, {target})"
            elif isinstance(time_value, QuaProgram.AnalogTimeDivision.Accumulated):
                self._out = f"{demod_name}. accumulated({common_args}, {time_value.samplesPerChunk}, {target})"
            elif isinstance(time_value, QuaProgram.AnalogTimeDivision.MovingWindow):
                self._out = f"{demod_name}.moving_window({common_args}, {time_value.samplesPerChunk}, {time_value.chunksPerWindow}, {target})"
            else:
                raise Exception(f"Unsupported analog process target {time_name}")
        else:
            raise Exception(f"Unsupported analog process target {target_name}")

    def visit_analog_measurement_dual_demod_integration(
        self, node: QuaProgram.AnalogMeasureProcess.DualDemodIntegration
    ) -> None:
        self._common_dual_measurement(node, "dual_demod")

    def visit_analog_measurement_dual_bare_integration(
        self, node: QuaProgram.AnalogMeasureProcess.DualBareIntegration
    ) -> None:
        self._common_dual_measurement(node, "dual_integration")

    def _common_time_tagging(
        self,
        node: Union[
            QuaProgram.AnalogMeasureProcess.RawTimeTagging,
            QuaProgram.AnalogMeasureProcess.HighResTimeTagging,
            QuaProgram.DigitalMeasureProcess.RawTimeTagging,
        ],
        demod_name: str,
    ) -> None:
        target = ExpressionSerializingVisitor(self._visitor).serialize(node.target)
        target_len = ExpressionSerializingVisitor(self._visitor).serialize(node.targetLen)
        max_time = node.maxTime
        element_output = node.elementOutput
        self._out = f'time_tagging.{demod_name}({target}, {max_time}, {target_len}, "{element_output}")'

    def visit_analog_measurement_raw_time_tagging(self, node: QuaProgram.AnalogMeasureProcess.RawTimeTagging) -> None:
        self._common_time_tagging(node, "analog")

    def visit_analog_measurement_high_res_time_tagging(
        self, node: QuaProgram.AnalogMeasureProcess.HighResTimeTagging
    ) -> None:
        self._common_time_tagging(node, "high_res")

    def visit_digital_measurement_counting(self, node: QuaProgram.DigitalMeasureProcess.Counting) -> None:
        element_outputs = []
        for element_output in node.elementOutputs:
            element_outputs.append(f'"{element_output}"')
        element_outputs_str = ",".join(element_outputs)
        target = ExpressionSerializingVisitor(self._visitor).serialize(node.target)
        max_time = node.maxTime
        self._out = f"counting.digital({target}, {max_time}, ({element_outputs_str}))"

    def visit_digital_measurement_raw_time_tagging(self, node: QuaProgram.DigitalMeasureProcess.RawTimeTagging) -> None:
        self._common_time_tagging(node, "digital")

    def visit_binary_expression(self, node: QuaProgram.BinaryExpression) -> None:
        left = ExpressionSerializingVisitor(self._visitor).serialize(node.left)
        right = ExpressionSerializingVisitor(self._visitor).serialize(node.right)
        sop = node.op
        mapping = {
            QuaProgram.BinaryExpression.BinaryOperator.ADD: "+",
            QuaProgram.BinaryExpression.BinaryOperator.SUB: "-",
            QuaProgram.BinaryExpression.BinaryOperator.GT: ">",
            QuaProgram.BinaryExpression.BinaryOperator.LT: "<",
            QuaProgram.BinaryExpression.BinaryOperator.LET: "<=",
            QuaProgram.BinaryExpression.BinaryOperator.GET: ">=",
            QuaProgram.BinaryExpression.BinaryOperator.EQ: "==",
            QuaProgram.BinaryExpression.BinaryOperator.MULT: "*",
            QuaProgram.BinaryExpression.BinaryOperator.DIV: "/",
            QuaProgram.BinaryExpression.BinaryOperator.OR: "|",
            QuaProgram.BinaryExpression.BinaryOperator.AND: "&",
            QuaProgram.BinaryExpression.BinaryOperator.XOR: "^",
            QuaProgram.BinaryExpression.BinaryOperator.SHL: "<<",
            QuaProgram.BinaryExpression.BinaryOperator.SHR: ">>",
        }
        if sop in mapping:
            op = mapping[sop]
        else:
            raise Exception(f"Unsupported operator {sop}")
        self._out = f"({left}{op}{right})"

    def visit_global_var_reference(self, node: QuaProgram.GlobalVarRefExpression) -> None:
        bits = ", ".join([str(b) for b in node.bits])
        if node.operation == QuaProgram.GlobalVarOperation.xor:
            self._out = f"global_var_xor({bits})"
        elif node.operation == QuaProgram.GlobalVarOperation.read:
            self._out = f"global_var_read({bits})"
        elif node.operation == QuaProgram.GlobalVarOperation.read_shift:
            self._out = f"global_var_read({bits}, shift=True)"

    def serialize(self, node: Node) -> str:
        self.visit(node)
        return self._out
